#![allow(dead_code)]

mod engine;

use engine::{Component, System};

#[derive(Component)]
struct Sprite {
    sprite: engine::Sprite,
}

#[derive(Component, Default, Clone)]
struct RigidBody {
    pos: (f64, f64),
    vel: (f64, f64),
    gravity: bool,
    cowision: bool,
}

struct VelocitySystem;
impl System for VelocitySystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id);
            body.pos.0 += body.vel.0 * delta;
            body.pos.1 += body.vel.1 * delta;
        }
        Ok(())
    }
}

#[derive(Component, Clone)]
struct Rect {
    width: f64,
    height: f64,
}

impl Rect {
    pub fn rect_collides(&self, pos: (f64, f64), other: &Rect, other_pos: (f64, f64)) -> bool {
        pos.0 + self.width > other_pos.0
            && pos.0 <= other_pos.0 + other.width
            && pos.1 + self.height > other_pos.1
            && pos.1 <= other_pos.1 + other.height
    }

    pub fn point_collides(&self, pos: (f64, f64), point: (f64, f64)) -> bool {
        pos.0 + self.width > point.0
            && pos.0 <= point.0
            && pos.1 + self.height > point.1
            && pos.1 <= point.1
    }
}
//
// enum CollisionGroup {
//     Player,
//     World,
// }
//
// impl CollisionGroup {
//     pub fn should_collide(&self, other: &Self) -> bool {
//         match (self, other) {
//             (CollisionGroup::Player, CollisionGroup::Player) => todo!(),
//             (CollisionGroup::Player, CollisionGroup::World) => todo!(),
//             (CollisionGroup::World, CollisionGroup::Player) => todo!(),
//             (CollisionGroup::World, CollisionGroup::World) => todo!(),
//         }
//     }
// }
//

#[derive(Component, Clone, Default)]
struct Collider {
    resolve: bool,
}

fn horizontal_line_intersect(p0: (f64, f64), vel: (f64, f64), line_y: f64) -> (f64, f64) {
    // y = ax + b
    let a = vel.1 / vel.0;
    let b = p0.1 - p0.0 * a;

    let x = (line_y - b) / a;
    (x, line_y)
}

fn vertical_line_intersect(p0: (f64, f64), vel: (f64, f64), line_x: f64) -> (f64, f64) {
    // y = ax + b
    let a = vel.1 / vel.0;
    let b = p0.1 - p0.0 * a;

    let y = a * line_x + b;
    (line_x, y)
}

enum Surface {
    Top,
    Bottom,
    Left,
    Right,
}

fn point_distance(a: (f64, f64), b: (f64, f64)) -> f64 {
    ((a.0 - b.0).abs().powi(2) + (a.1 - b.1).abs().powi(2)).sqrt()
}

fn closest_surface_for_point_and_rectangle_and_your_mom(
    p0: (f64, f64),
    vel: (f64, f64),
    rect_pos: (f64, f64),
    rect: &Rect,
) -> Option<(f64, Surface)> {
    [
        (horizontal_line_intersect(p0, vel, rect_pos.1), Surface::Top),
        (
            horizontal_line_intersect(p0, vel, rect_pos.1 + rect.height),
            Surface::Bottom,
        ),
        (vertical_line_intersect(p0, vel, rect_pos.0), Surface::Left),
        (
            horizontal_line_intersect(p0, vel, rect_pos.0 + rect.width),
            Surface::Right,
        ),
    ]
    .into_iter()
    .filter(|(point, _)| rect.point_collides(rect_pos, *point))
    .map(|(point, surface)| (point_distance(p0, point), surface))
    .min_by(|(dist_a, _), (dist_b, _)| dist_a.total_cmp(&dist_b))
}

struct CollisionSystem;
impl System for CollisionSystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, RigidBody, Rect, Collider) {
            let collider = ctx.entity_component::<Collider>(id).clone();
            if !collider.resolve {
                continue;
            }
            let body = ctx.entity_component::<RigidBody>(id).clone();
            let rect = ctx.entity_component::<Rect>(id).clone();
            for other_id in query!(ctx, RigidBody, Rect, Collider) {
                if id == other_id {
                    continue;
                }
                let other_rect = ctx.entity_component::<Rect>(id).clone();
                let other_body = ctx.entity_component::<RigidBody>(id).clone();
                if rect.rect_collides(body.pos, &other_rect, other_body.pos) {
                    println!("collider vi?");
                    let last_pos = (
                        body.pos.0 - body.vel.0 * delta,
                        body.pos.1 - body.vel.1 * delta,
                    );
                    let closest_surface = [
                        (last_pos.0, last_pos.1),
                        (last_pos.0, last_pos.1 + rect.height),
                        (last_pos.0 + rect.width, last_pos.1),
                        (last_pos.0 + rect.width, last_pos.1 + rect.height),
                    ]
                    .into_iter()
                    .map(|p0| {
                        closest_surface_for_point_and_rectangle_and_your_mom(
                            p0,
                            body.vel,
                            other_body.pos,
                            &other_rect,
                        )
                    })
                    .filter_map(std::convert::identity)
                    .min_by(|(dist_a, _), (dist_b, _)| dist_a.total_cmp(&dist_b))
                    .map(|(_, surface)| surface)
                    .ok_or_else(|| "we already checked if collision happens")?;

                    let body = ctx.entity_component::<RigidBody>(id);
                    match closest_surface {
                        Surface::Top => {
                            body.vel.1 = 0.0;
                            body.pos.1 = other_body.pos.1 - rect.height;
                        }
                        Surface::Bottom => {
                            body.vel.1 = 0.0;
                            body.pos.1 = other_body.pos.1 + other_rect.height;
                        }
                        Surface::Left => {
                            body.vel.0 = 0.0;
                            body.pos.0 = other_body.pos.0 + other_rect.width;
                        }
                        Surface::Right => {
                            body.vel.0 = 0.0;
                            body.pos.0 = other_body.pos.0 - rect.width;
                        }
                    }
                }
            }
        }
        Ok(())
    }
}
struct GravitySystem;
impl System for GravitySystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id);
            if !body.gravity {
                continue;
            }
            body.vel.1 = if body.vel.1 < 800.0 {
                body.vel.1 + 400.0 * delta
            } else {
                body.vel.1
            };
        }
        Ok(())
    }
}

#[derive(Component)]
struct Cloud;

struct CloudSystem;
impl System for CloudSystem {
    fn on_update(&self, ctx: &mut engine::Context, delta: f64) -> Result<(), engine::Error> {
        let cloud_amount = ctx.entities_with_component::<Cloud>().len();
        if cloud_amount < 1 {
            let cloud = ctx.load_sprite("textures/clouds.png").unwrap();
            spawn!(
                ctx,
                Cloud,
                Sprite { sprite: cloud },
                RigidBody {
                    pos: (-100.0, 150.0),
                    ..Default::default()
                },
            );
        }

        for id in query!(ctx, Cloud, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id);
            body.vel.0 = if body.vel.0 < 200.0 {
                body.vel.0 + 200.0 * delta
            } else {
                body.vel.0
            };
        }

        for id in query!(ctx, Cloud, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id);
            if body.pos.0 > 1400.0 {
                ctx.despawn(id);
            }
        }
        Ok(())
    }
}

struct SpriteRenderer;
impl System for SpriteRenderer {
    fn on_update(&self, ctx: &mut engine::Context, _delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, Sprite, RigidBody) {
            let body = ctx.entity_component::<RigidBody>(id).clone();
            let sprite = ctx.entity_component::<Sprite>(id).sprite;

            ctx.draw_sprite(sprite, body.pos.0 as i32, body.pos.1 as i32)?;
        }
        Ok(())
    }
}

#[derive(Component)]
struct PlayerMovement;

struct PlayerMovementSystem;
impl System for PlayerMovementSystem {
    fn on_update(&self, ctx: &mut engine::Context, _delta: f64) -> Result<(), engine::Error> {
        for id in query!(ctx, PlayerMovement, RigidBody) {
            let d_down = ctx.key_pressed(engine::Keycode::D);
            let a_down = ctx.key_pressed(engine::Keycode::A);
            let body = ctx.entity_component::<RigidBody>(id);
            body.vel.0 = if d_down && !a_down {
                400.0
            } else if !d_down && a_down {
                -400.0
            } else {
                0.0
            };
        }
        Ok(())
    }
}

fn main() {
    let mut game = engine::Game::new().unwrap();

    let mut context = game.context();
    context.add_system(VelocitySystem);
    context.add_system(CollisionSystem);
    context.add_system(SpriteRenderer);
    context.add_system(GravitySystem);
    context.add_system(PlayerMovementSystem);
    context.add_system(CloudSystem);
    let player = context.load_sprite("textures/player.png").unwrap();
    let background = context.load_sprite("textures/literally_dprk.png").unwrap();

    spawn!(
        &mut context,
        Sprite { sprite: background },
        RigidBody::default(),
    );

    spawn!(
        &mut context,
        Sprite { sprite: player },
        RigidBody {
            pos: (400.0, 400.0),
            gravity: true,
            ..Default::default()
        },
        Rect {
            width: 256.0,
            height: 256.0
        },
        Collider { resolve: true },
        PlayerMovement,
    );

    spawn!(
        &mut context,
        RigidBody {
            pos: (0.0, 650.0),
            ..Default::default()
        },
        Rect {
            width: 1000.0,
            height: 25.0
        },
        Collider { resolve: false },
    );

    game.run();
}
